<?php

declare(strict_types=1);

namespace Cin\Bistro5;

use DateTimeImmutable;
use JTL\Shop;
use PDO;
use stdClass;

final class BistroService
{
  public const STORE_INTERVAL = 1;

  public const TABLE_CLOSINGS = 'xplugin_cin_bistro_closings';

  public const TABLE_OPENINGS = 'xplugin_cin_bistro_openings';

  public const TABLE_ORDERNUMBER = 'xplugin_cin_bistro_order_number';

  private const RND_CHARSET = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ';

  private static $p;

  private static $pdo;

  public static function DayToggle($data): void
  {
    foreach ($data as $k => $b) {
      $db = Shop::Container()->getDB();
      if ( ! $db->update(self::TABLE_OPENINGS, 'day', 'x'.mb_substr($k, 0, 2), (object) ['active' => ($b ? 1 : 0), 'close' => 'now()'])) {
        $db->insert(self::TABLE_OPENINGS, (object) ['day' => 'x'.mb_substr($k, 0, 2), 'active' => ($b ? 1 : 0)]);
      }
    }
  }

  public static function DeleteClosings(): void
  {
    $db = Shop::Container()->getDB();

    $prep = $db->getPDO()->prepare('DELETE FROM '.self::TABLE_CLOSINGS);
    $prep->execute([]);
  }

  public static function DeleteLeaf(array $data): array
  {
    $db = Shop::Container()->getDB();

    $db->delete(self::TABLE_OPENINGS, 'id', $data['id']);

    $prep = $db->getPDO()->prepare(sprintf('SELECT COUNT(*) count FROM %s WHERE day = ?', self::TABLE_OPENINGS));
    $prep->execute([$data['day']]);
    $rslt = $prep->fetch(PDO::FETCH_ASSOC);

    if (0 === $rslt['count']) {
      $db->update(self::TABLE_OPENINGS, 'day', 'x'.mb_substr($data['day'], 0, 2), (object) ['active' => 0]);
    }

    return $rslt;
  }

  public static function ExtendBranch(string $branchID): array
  {
    $db  = Shop::Container()->getDB();
    $pdo = $db->getPDO();

    $db->insert(self::TABLE_OPENINGS, (object) ['day' => $branchID]);

    $stmt = sprintf('SELECT `id`, `open`, `close` FROM %s WHERE `day` = ?', self::TABLE_OPENINGS);
    $prep = $pdo->prepare($stmt);
    $prep->execute([$branchID]);
    $list = $prep->fetchAll();

    return $list ?? [];
  }

  /**
   * Get an array with foodtimes and there count for current day.
   */
  public static function Foodtimes(): array
  {
    $prep = self::PDO()->prepare("SELECT
      #b.kBestellung,
      bat.cValue foodtime,
      #bao.cValue foodorder,
      COUNT(bat.cValue) count
      FROM `tbestellung` b
      LEFT JOIN `tbestellattribut` bat ON bat.kBestellung = b.kBestellung AND bat.cName = 'cin_foodtime'
      #LEFT JOIN `tbestellattribut` bao ON bao.kBestellung = b.kBestellung AND bao.cName = 'cin_foodorder'
      WHERE DATE(b.dErstellt) = DATE(now())
      AND bat.cValue IS NOT NULL
      GROUP BY bat.cValue");

    $prep->execute();

    $ft      = $prep->fetchAll(PDO::FETCH_UNIQUE);
    $cleaned = [];

    foreach ($ft as $key => $value) {
      $ts                = date('H:i:s', $key);
      $dz_time           = strtotime("2000-01-01 {$ts}");
      $cleaned[$dz_time] = $value;
    }

    return $cleaned;
  }

  public static function GetClosings()
  {
    $db = Shop::Container()->getDB();

    return (array) ($db->selectSingleRow(self::TABLE_CLOSINGS, [], []) ?? []);
  }

  public static function InitialDayToggles()
  {
    $db = Shop::Container()->getDB();

    $days = [
      'mon' => false,
      'tue' => false,
      'wed' => false,
      'thu' => false,
      'fri' => false,
      'sat' => false,
      'sun' => false
    ];

    array_walk($days, static function (&$val, $key, $db): void {
      $dbdata = $db->select(self::TABLE_OPENINGS, 'day', 'x'.mb_substr($key, 0, 2)) ?? ['active' => false];
      $val    = (bool) (int) ((array) $dbdata)['active'];
    }, $db);

    return $days;
  }

  public static function IsDayInClosingHours(): bool
  {
    $cHours = self::GetClosings();
    if (empty($cHours)) {
      return false;
    }

    $closingHours = [];
    if (isset($cHours['from'])) {
      $closingHours[] = $cHours;
    } else {
      $closingHours = $cHours;
    }

    $dNow = DateTimeImmutable::createFromFormat('Y-m-d H:i:s', date('Y-m-d H:i:s'));

    foreach ($closingHours as $closing) {
      $dFrom = DateTimeImmutable::createFromFormat('Y-m-d H:i:s', $closing['from']);
      $dTo   = DateTimeImmutable::createFromFormat('Y-m-d H:i:s', mb_substr($closing['to'], 0, 11).'23:59:59');

      if ($dNow >= $dFrom && $dNow <= $dTo) {
        return true;
      }
    }

    return false;
  }

  public static function IsGenerallyOpened(): bool
  {
    $curday = mb_strtolower(date('D'));
    $db     = Shop::Container()->getDB();
    $rslt   = (array) $db->selectSingleRow(self::TABLE_OPENINGS, 'day', 'x'.mb_substr($curday, 0, 2), null, null, null, null, false, 'active') ?? ['active' => false];

    return (bool) $rslt['active'];
  }

  public static function IsOpened()
  {
    $timesOfDay = self::TimesOfDay();
    if (empty($timesOfDay)) {
      return false;
    }

    $dNow = DateTimeImmutable::createFromFormat('Y-m-d H:i:s', date('Y-m-d H:i:s'));

    foreach ($timesOfDay as $times) {
      $times  = (array) $times;
      $dOpen  = DateTimeImmutable::createFromFormat('Y-m-d H:i:s', date('Y-m-d ').$times['open']);
      $dClose = DateTimeImmutable::createFromFormat('Y-m-d H:i:s', date('Y-m-d ').$times['close']);

      if ($dNow >= $dOpen && $dNow <= $dClose) {
        return true;
      }
    }

    return false;
  }

  public static function LoadBranches(): array
  {
    $db  = Shop::Container()->getDB();
    $pdo = $db->getPDO();

    $stmt = sprintf("SELECT week.day, o.id, o.open, o.close
    FROM (
      SELECT 'mon' AS day, '0' AS id UNION ALL
      SELECT 'tue', '1' UNION ALL
      SELECT 'wed', '2' UNION ALL
      SELECT 'thu', '3' UNION ALL
      SELECT 'fri', '4' UNION ALL
      SELECT 'sat', '5' UNION ALL
      SELECT 'sun', '6' ) AS week
    LEFT JOIN %s o ON week.day = o.day
    ORDER BY week.id", self::TABLE_OPENINGS);
    $prep = $pdo->prepare($stmt);
    $prep->execute([]);
    $list = $prep->fetchAll(PDO::FETCH_GROUP);

    return $list ?? [];
  }

  public static function NextClosing(): array
  {
    return self::NextTiming('close');
  }

  public static function NextOpening()
  {
    return self::NextTiming('open');
  }

  public static function NoticeBistroId(): void
  {
    $db  = Shop::Container()->getDB();
    $pdo = $db->getPDO();

    $stmt = sprintf('SELECT `order_id` FROM `%s` WHERE `order_id` = ? LIMIT 1', self::TABLE_ORDERNUMBER);

    do {
      $oid  = self::generateRandomString();
      $prep = $pdo->prepare($stmt);
      $prep->execute([$oid]);
      $rslt = $prep->fetchColumn();
    } while ($rslt);

    $db->insert(self::TABLE_ORDERNUMBER, (object) ['order_id' => $oid]);

    $attr         = new stdClass();
    $attr->cName  = 'cin_foodorder';
    $attr->cValue = $oid;

    $_SESSION['Warenkorb']->OrderAttributes[] = $attr;
  }

  public static function NoticeBistroTime(): void
  {
    if (isset($_POST['bistrozeit']) && -1 !== $_POST['bistrozeit'] && '-1' !== $_POST['bistrozeit']) {
      $attr         = new stdClass();
      $attr->cName  = 'cin_foodtime';
      $attr->cValue = strtotime(date('Y-m-d ').$_POST['bistrozeit']);

      $_SESSION['Warenkorb']->OrderAttributes[] = $attr;
    }
  }

  public static function PossibleDeliveryTimes(): array
  {
    $curdayTimes       = self::TimesOfDay();
    $curtime           = date('Y-m-d H:i:s', mktime((int) date('H'), (int) date('i'), 0, 1, 1, 2000));
    $finalDateTime     = date('Y-m-d H:i:s', strtotime('2000-01-01 + 24 HOURS'));
    $additionalMinutes = (int) self::P()->cfg()['time_clocking'];
    $delivery_times    = [];
    $clockLimit        = (int) self::P()->cfg()['clocking_limit'];
    $foodtimes         = 0 < $clockLimit ? self::Foodtimes() : [];
    $increaseClocking  = static function (string $clocking, int $frequence): string {
      return date('Y-m-d H:i:s', strtotime("{$clocking} + {$frequence} MINUTES"));
    };

    foreach ($curdayTimes as $times) {
      $from         = $times->open;
      $to           = date('H:i:s', strtotime("2000-01-01 {$times->close} - {$additionalMinutes} MINUTES"));
      $nextDateTime = date('Y-m-d H:i:s', strtotime('2000-01-01'));

      while (strtotime($nextDateTime) <= strtotime($finalDateTime)) {
        if (strtotime($nextDateTime) < strtotime("2000-01-01 {$from}")
          || strtotime($nextDateTime) < strtotime("{$curtime} + {$additionalMinutes} MINUTES")) {
          $nextDateTime = $increaseClocking($nextDateTime, $additionalMinutes);

          continue;
        }
        if (strtotime($nextDateTime) > strtotime('2000-01-01 '.$to)) {
          break;
        }

        if ($clockLimit && 'hide' === self::P()->cfg()['clock_blocking'] && isset($foodtimes[strtotime($nextDateTime)]) && $clockLimit <= $foodtimes[strtotime($nextDateTime)]->count) {
          $nextDateTime = $increaseClocking($nextDateTime, $additionalMinutes);

          continue;
        }

        $blocked = ($clockLimit && isset($foodtimes[strtotime($nextDateTime)]) && $clockLimit <= $foodtimes[strtotime($nextDateTime)]->count);

        $delivery_times[] = [
          'sel' => false,
          'txt' => date('H:i', strtotime($nextDateTime)),
          'val' => $blocked ? 'blocked' : date('H:i', strtotime($nextDateTime))
        ];
        $nextDateTime = $increaseClocking($nextDateTime, $additionalMinutes);
      }
    }

    sort($delivery_times);

    if ('on' === self::P()->cfg()['time_quickly']) {
      array_unshift($delivery_times, ['sel' => false, 'txt' => __('bistro_fast_as_possible'), 'val' => -1]);
    }

    return $delivery_times;
  }

  public static function RemoveOldIDs(): void
  {
    $pdo  = self::PDO();
    $prep = $pdo->prepare(sprintf('DELETE FROM %s WHERE DATE_ADD(order_date ,INTERVAL %d DAY) < NOW()', self::TABLE_ORDERNUMBER, self::STORE_INTERVAL));
    $prep->execute();
  }

  public static function Test($data): void
  {
    $db = Shop::Container()->getDB();

    // return self::GetCompareableTimestampCurrent();

    // return self::GetWeekDays(true);

    // return self::NextOpening();

    // return self::NextClosing();

    // return self::TimesOfDay();

    // return self::IsOpened();

    // return $db->selectSingleRow(self::TABLE_OPENINGS, 'day', 'x'.substr('thu', 0, 2), null, null, null, null, false, 'active')??['active' => false];

    // return self::IsDayInClosingHours();

    // return self::LoadBranches();

    // return self::GetClosings();

    // return $db->delet(self::TABLE_CLOSINGS, [], [], true);

    // return $db->select(self::TABLE_CLOSINGS, ['id'], ['2'], null, null, null, null, false, 'id');

    // $dbdata = $db->selectAll(self::TABLE_OPENINGS, 'day', 'mon', 'open, close');
    // return array_filter($dbdata, function ($row) {
    //   return (\DateTime::createFromFormat('H:i:s', $row->open) && \DateTime::createFromFormat('H:i:s', $row->close));
    // });
  }

  public static function TimesOfDay(?string $curday = null): array
  {
    if (self::IsDayInClosingHours()) {
      return [];
    }

    $curday = ($curday ?: mb_strtolower(date('D')));

    $stmt = sprintf('SELECT active FROM %s WHERE `day` = :day', self::TABLE_OPENINGS);
    $prep = self::PDO()->prepare($stmt);
    $prep->execute([
      ':day' => 'x'.mb_substr($curday, 0, 2)
    ]);
    if ( ! (bool) $prep->fetchColumn()) {
      return [];
    }

    $stmt = sprintf('SELECT `open`, `close` FROM %s WHERE `day` = :day', self::TABLE_OPENINGS);
    $prep = self::PDO()->prepare($stmt);
    $prep->execute([
      ':day' => $curday
    ]);

    return array_filter($prep->fetchAll(), static fn ($row) => DateTimeImmutable::createFromFormat('H:i:s', $row->open ?? '') && DateTimeImmutable::createFromFormat('H:i:s', $row->close ?? ''));
  }

  public static function RowTemplate()
  {
    // 0 $branchId
    // 1 $branchLeafId
    // 2 $branchLeaf[0]
    // 3 $branchLeaf[1]
    // 4 $branchLeaf[2]

    return '
    <div class="row {0}_{1}">
    <input type="hidden" id="delivery_time_i_{0}_{1}" name="delivery_time_i_{0}_{1}" value="{2}"/>
    <input class="col form-control delivery_time_o" id="delivery_time_o_{0}_{1}" name="delivery_time_o_{0}_{1}" type="time" value="{3}"/>
    <input class="col form-control delivery_time_c" id="delivery_time_c_{0}_{1}" name="delivery_time_c_{0}_{1}" type="time" value="{4}"/>
    <button class="col fal fa-trash btn btn-danger btn-delete" type="button" disabled></button>
    </div>';
  }

  /**
   * Insert or Update a time for specified day.
   *
   * @param array $data Expects <ul><li>id (database id or null for new),</li><li>day (3 digits),</li><li>time (12:00),</li><li>opentime (boolean as flag for is open or close time)</li></ul>
   */
  public static function UpdateTime(array $data): bool
  {
    $db = Shop::Container()->getDB();

    if (5 === mb_strlen($data['time'])) {
      $data['time'] = $data['time'].':00';
    }
    $timeIsValid = DateTimeImmutable::createFromFormat('H:i:s', $data['time']);

    if ( ! $timeIsValid) {
      return false;
    }

    $isUpdate = (null !== $data['id'] && (int) $data['id'] > 0);

    if ($isUpdate) {
      $db->update(self::TABLE_OPENINGS, 'id', $data['id'], (object) [($data['opentime'] ? 'open' : 'close') => $data['time']]);
    }
    // else {
    //   $db->insert(self::TABLE_OPENINGS, (object)['day' => $data['day'], 'open' => ($data['opentime'] ? $data['time'] : '00:00:00'), 'close' => ($data['opentime'] ? '00:00:00' : $data['time'])]);
    // }

    return true;
  }

  public static function UpdateClosings(array $data): array
  {
    $failures = [];

    if ( ! DateTimeImmutable::createFromFormat('Y-m-d', $data['from'])) {
      $failures['from'] = 1;
    }
    if ( ! DateTimeImmutable::createFromFormat('Y-m-d', $data['to'])) {
      $failures['to'] = 1;
    }
    if (DateTimeImmutable::createFromFormat('Y-m-d', $data['from']) >= DateTimeImmutable::createFromFormat('Y-m-d', $data['to'])) {
      $failures['from'] = $failures['to'] = 1;
    }
    if ( ! empty($failures)) {
      return $failures;
    }

    $db = Shop::Container()->getDB();

    $rslt = $db->select(self::TABLE_CLOSINGS, [], [], null, null, null, null, false, 'id');

    if (null === $rslt) {
      $db->insert(self::TABLE_CLOSINGS, (object) $data);
    } else {
      $db->update(self::TABLE_CLOSINGS, 'id', $rslt->id, (object) $data);
    }

    return $failures;
  }

  /**
   * Create a random string.
   *
   * @param int $length Length of the new string. Default is 4.
   */
  public static function generateRandomString(int $length = 4): string
  {
    $charactersLength = mb_strlen(self::RND_CHARSET);
    $randomString     = '';
    for ($i = 0; $i < $length; ++$i) {
      $randomString .= self::RND_CHARSET[random_int(0, $charactersLength - 1)];
    }

    return $randomString;
  }

  /**
   * Timestamp with current time but date of 2000-01-01 for comparing with other times.
   */
  private static function GetCompareableTimestampCurrent(): int
  {
    return strtotime('2000-01-01 '.date('H:i:s'));
  }

  /**
   * Return an array with shorted das of week (mon - sun).
   *
   * @param bool $sorted Sort array beginning with current day of week
   */
  private static function GetWeekDays(bool $sorted = false): array
  {
    $days = ['mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun'];
    if ( ! $sorted) {
      return $days;
    }

    $curday = mb_strtolower(date('D'));
    while ($days[0] !== $curday) {
      $day    = array_shift($days);
      $days[] = $day;
    }

    return $days;
  }

  private static function NextTiming(string $_oc): array
  {
    $days = self::GetWeekDays(true);

    $timeNow = self::GetCompareableTimestampCurrent();
    $rslt    = [];

    foreach ($days as $idx => $day) {
      $times = self::TimesOfDay($day);
      foreach ($times as $timeRow) {
        $at = strtotime("2000-01-01 {$timeRow->{$_oc}} + {$idx} DAYS");
        if ($timeNow < $at) {
          $rslt[$_oc.'At'] = $timeRow->{$_oc};
          $rslt['onDay']   = $day;
          $rslt['nth']     = $idx;

          return $rslt;
        }
      }
    }

    return $rslt;
  }

  private static function P(): Plugin
  {
    return self::$p ?? (self::$p = new Plugin());
  }

  private static function PDO()
  {
    return self::$pdo ?? (self::$pdo = self::P()->getPDO());
  }
}
